import autofit as af


class CosmologyException(Exception):
    pass


class GalaxyException(Exception):
    pass


class PlottingException(Exception):
    pass


class UnitsException(Exception):
    pass


class PriorException(Exception):
    pass
